@@ -132,7 +132,7 @@ type ('a, 'b, 'c) stream = {
   mutable decoder : ('b, Avcodec.decode) Avcodec.codec option;
 }
 
-type media_type = MT_audio | MT_video | MT_subtitle
+type media_type = MT_audio | MT_video | MT_data | MT_subtitle
 
 let mk_stream container index = { container; index; decoder = None }
 
@@ -163,6 +163,7 @@ let get_streams input media_type =
 let get_audio_streams input = get_streams input MT_audio
 let get_video_streams input = get_streams input MT_video
 let get_subtitle_streams input = get_streams input MT_subtitle
+let get_data_streams input = get_streams input MT_data
 let set_decoder s decoder = s.decoder <- Some decoder
 
 external _find_best_stream : input container -> media_type -> int
@@ -190,32 +191,43 @@ type input_result =
   | `Video_packet of int * video Avcodec.Packet.t
   | `Video_frame of int * video frame
   | `Subtitle_packet of int * subtitle Avcodec.Packet.t
-  | `Subtitle_frame of int * subtitle frame ]
+  | `Subtitle_frame of int * subtitle frame
+  | `Data_packet of int * [ `Data ] Avcodec.Packet.t ]
 
 (** Reads the selected streams if any or all streams otherwise. *)
 external read_input :
-  (int * ('a, Avcodec.decode) Avcodec.codec option) array ->
+  (int * Avutil.media_type) array ->
   (int * ('a, Avcodec.decode) Avcodec.codec option) array ->
   input container ->
   input_result = "ocaml_av_read_input"
 
-let _get input =
+let _get_packet media_type input =
+  List.map (fun { index; container; _ } ->
+      if container != input then
+        raise (Failure "Inconsistent stream and input!");
+      (index, media_type))
+
+let _get_frame input =
   List.map (fun { index; container; decoder } ->
       if container != input then
         raise (Failure "Inconsistent stream and input!");
       (index, Obj.magic decoder))
 
 let read_input ?(audio_packet = []) ?(audio_frame = []) ?(video_packet = [])
-    ?(video_frame = []) ?(subtitle_packet = []) ?(subtitle_frame = []) input =
+    ?(video_frame = []) ?(subtitle_packet = []) ?(subtitle_frame = [])
+    ?(data_packet = []) input =
   let packet =
     Array.of_list
-      (_get input audio_packet @ _get input video_packet
-     @ _get input subtitle_packet)
+      (_get_packet `Audio input audio_packet
+      @ _get_packet `Video input video_packet
+      @ _get_packet `Subtitle input subtitle_packet
+      @ _get_packet `Data input data_packet)
   in
   let frame =
     Array.of_list
-      (_get input audio_frame @ _get input video_frame
-     @ _get input subtitle_frame)
+      (_get_frame input audio_frame
+      @ _get_frame input video_frame
+      @ _get_frame input subtitle_frame)
   in
   read_input packet frame input
 
@@ -279,12 +291,24 @@ let set_output_metadata o tags = _set_metadata o (-1) (Array.of_list tags)
 let set_metadata s tags = _set_metadata s.container s.index (Array.of_list tags)
 let get_output s = s.container
 
-external new_stream_copy : output container -> _ Avcodec.params -> int
-  = "ocaml_av_new_stream_copy"
+type uninitialized_stream_copy = output container * int
+
+external new_uninitialized_stream_copy : output container -> int
+  = "ocaml_av_new_uninitialized_stream_copy"
+
+let new_uninitialized_stream_copy container =
+  (container, new_uninitialized_stream_copy container)
+
+external initialize_stream_copy :
+  output container -> int -> _ Avcodec.params -> unit
+  = "ocaml_av_initialize_stream_copy"
+
+let initialize_stream_copy ~params (container, index) =
+  initialize_stream_copy container index params;
+  mk_stream container index
 
 let new_stream_copy ~params container =
-  let ret = new_stream_copy container params in
-  mk_stream container ret
+  initialize_stream_copy ~params (new_uninitialized_stream_copy container)
 
 external new_audio_stream :
   _ container ->
@@ -357,6 +381,14 @@ let new_subtitle_stream ?opts ~time_base ~codec container =
   filter_opts unused opts;
   mk_stream container ret
 
+external new_data_stream :
+  _ container -> Avcodec.Unknown.id -> Avutil.rational -> int
+  = "ocaml_av_new_data_stream"
+
+let new_data_stream ~time_base ~codec container =
+  let ret = new_data_stream container codec time_base in
+  mk_stream container ret
+
 external codec_attr : _ stream -> string option = "ocaml_av_codec_attr"
 external bitrate : _ stream -> int option = "ocaml_av_stream_bitrate"
 
