@@ -1,5 +1,7 @@
 #include <string.h>
 
+#define CAML_NAME_SPACE 1
+
 #include <caml/alloc.h>
 #include <caml/bigarray.h>
 #include <caml/callback.h>
@@ -54,6 +56,7 @@ typedef struct av_t {
   value control_message_callback;
   int is_input;
   value interrupt_cb;
+  int closed;
 
   // input
   int end_of_file;
@@ -67,7 +70,13 @@ typedef struct av_t {
   int custom_io;
 } av_t;
 
-#define Av_val(v) (*(av_t **)Data_custom_val(v))
+#define Av_base_val(v) (*(av_t **)Data_custom_val(v))
+static inline av_t *Av_val(value v) {
+  av_t *av = Av_base_val(v);
+  if (av->closed)
+    Fail("Container closed!");
+  return av;
+}
 
 /***** Stream handler *****/
 
@@ -77,7 +86,8 @@ typedef struct av_t {
 /**** Media Type ****/
 
 static const enum AVMediaType MEDIA_TYPES[] = {
-    AVMEDIA_TYPE_AUDIO, AVMEDIA_TYPE_VIDEO, AVMEDIA_TYPE_SUBTITLE};
+    AVMEDIA_TYPE_AUDIO, AVMEDIA_TYPE_VIDEO, AVMEDIA_TYPE_DATA,
+    AVMEDIA_TYPE_SUBTITLE};
 #define MEDIA_TYPES_LEN (sizeof(MEDIA_TYPES) / sizeof(enum AVMediaType))
 
 enum AVMediaType MediaType_val(value v) { return MEDIA_TYPES[Int_val(v)]; }
@@ -133,9 +143,11 @@ static void close_av(av_t *av) {
     caml_remove_generational_global_root(&av->interrupt_cb);
     av->interrupt_cb = Val_none;
   }
+
+  av->closed = 1;
 }
 
-static void finalize_av(value v) { free(Av_val(v)); }
+static void finalize_av(value v) { free(Av_base_val(v)); }
 
 static struct custom_operations av_ops = {
     "ocaml_av_context",       finalize_av,
@@ -545,7 +557,7 @@ CAMLprim value caml_av_input_io_finalise(value _avio) {
 
   // format_context is freed as part of close_av.
   av_free(avio->avio_context->buffer);
-  avio_context_free(avio->avio_context);
+  avio_context_free(&avio->avio_context);
 
   if (avio->read_cb)
     caml_remove_generational_global_root(&avio->read_cb);
@@ -632,6 +644,7 @@ static av_t *open_input(char *url, avioformat_const AVInputFormat *format,
     caml_raise_out_of_memory();
   }
 
+  av->closed = 0;
   av->is_input = 1;
   av->frames_pending = 0;
   av->streams = NULL;
@@ -737,7 +750,7 @@ CAMLprim value ocaml_av_open_input(value _url, value _format, value _interrupt,
 
   // allocate format context
   ans = caml_alloc_custom(&av_ops, sizeof(av_t *), 0, 1);
-  Av_val(ans) = av;
+  Av_base_val(ans) = av;
 
   ret = caml_alloc_tuple(2);
   Store_field(ret, 0, ans);
@@ -788,7 +801,7 @@ CAMLprim value ocaml_av_open_input_stream(value _avio, value _format,
 
   // allocate format context
   ans = caml_alloc_custom(&av_ops, sizeof(av_t *), 0, 1);
-  Av_val(ans) = av;
+  Av_base_val(ans) = av;
 
   ret = caml_alloc_tuple(2);
   Store_field(ret, 0, ans);
@@ -929,6 +942,7 @@ static stream_t *open_stream_index(av_t *av, int index, const AVCodec *dec) {
   err = avcodec_parameters_to_context(stream->codec_context, dec_param);
 
   if (err < 0) {
+    free(stream);
     ocaml_avutil_raise_error(err);
   }
 
@@ -937,8 +951,10 @@ static stream_t *open_stream_index(av_t *av, int index, const AVCodec *dec) {
   err = avcodec_open2(stream->codec_context, dec, NULL);
   caml_acquire_runtime_system();
 
-  if (err < 0)
+  if (err < 0) {
+    free(stream);
     ocaml_avutil_raise_error(err);
+  }
 
   return stream;
 }
@@ -1065,13 +1081,20 @@ CAMLprim value ocaml_av_read_input(value _packet, value _frame, value _av) {
       skip = 1;
       for (i = 0; i < Wosize_val(_packet); i++)
         if (Int_val(Field(Field(_packet, i), 0)) == packet->stream_index) {
-          _dec = Field(Field(_packet, i), 1);
           skip = 0;
         }
 
       for (i = 0; i < Wosize_val(_frame); i++)
         if (Int_val(Field(Field(_frame, i), 0)) == packet->stream_index) {
           _dec = Field(Field(_frame, i), 1);
+
+          if (_dec != Val_none) {
+            dec = AvCodec_val(Some_val(_dec));
+          }
+
+          if ((stream = streams[packet->stream_index]) == NULL)
+            stream = open_stream_index(av, packet->stream_index, dec);
+
           skip = 0;
         }
 
@@ -1080,13 +1103,6 @@ CAMLprim value ocaml_av_read_input(value _packet, value _frame, value _av) {
         continue;
       }
 
-      if (_dec != Val_none) {
-        dec = AvCodec_val(Some_val(_dec));
-      }
-
-      if ((stream = streams[packet->stream_index]) == NULL)
-        stream = open_stream_index(av, packet->stream_index, dec);
-
       for (i = 0; i < Wosize_val(_packet); i++) {
         if (Int_val(Field(Field(_packet, i), 0)) == packet->stream_index) {
           decoded_content = caml_alloc_tuple(2);
@@ -1095,14 +1111,16 @@ CAMLprim value ocaml_av_read_input(value _packet, value _frame, value _av) {
 
           ans = caml_alloc_tuple(2);
 
-          switch (
-              av->streams[packet->stream_index]->codec_context->codec_type) {
-          case AVMEDIA_TYPE_AUDIO:
+          switch (Field(Field(_packet, i), 1)) {
+          case PVV_Audio:
             Store_field(ans, 0, PVV_Audio_packet);
             break;
-          case AVMEDIA_TYPE_VIDEO:
+          case PVV_Video:
             Store_field(ans, 0, PVV_Video_packet);
             break;
+          case PVV_Data:
+            Store_field(ans, 0, PVV_Data_packet);
+            break;
           default:
             Store_field(ans, 0, PVV_Subtitle_packet);
             break;
@@ -1366,6 +1384,8 @@ static av_t *open_output(avioformat_const AVOutputFormat *format,
     caml_raise_out_of_memory();
   }
 
+  av->closed = 0;
+
   if (_interrupt != Val_none) {
     av->interrupt_cb = Some_val(_interrupt);
     caml_register_generational_global_root(&av->interrupt_cb);
@@ -1507,7 +1527,7 @@ CAMLprim value ocaml_av_open_output(value _interrupt, value _format,
 
   // allocate format context
   ans = caml_alloc_custom(&av_ops, sizeof(av_t *), 0, 1);
-  Av_val(ans) = av;
+  Av_base_val(ans) = av;
 
   ret = caml_alloc_tuple(2);
   Store_field(ret, 0, ans);
@@ -1554,7 +1574,7 @@ CAMLprim value ocaml_av_open_output_format(value _format, value _opts) {
 
   // allocate format context
   ans = caml_alloc_custom(&av_ops, sizeof(av_t *), 0, 1);
-  Av_val(ans) = av;
+  Av_base_val(ans) = av;
 
   ret = caml_alloc_tuple(2);
   Store_field(ret, 0, ans);
@@ -1602,7 +1622,7 @@ CAMLprim value ocaml_av_open_output_stream(value _format, value _avio,
 
   // allocate format context
   ans = caml_alloc_custom(&av_ops, sizeof(av_t *), 0, 1);
-  Av_val(ans) = av;
+  Av_base_val(ans) = av;
 
   ret = caml_alloc_tuple(2);
   Store_field(ret, 0, ans);
@@ -1633,6 +1653,7 @@ CAMLprim value ocaml_av_set_metadata(value _av, value _stream_index,
 
   int i, ret, len = Wosize_val(_tags);
 
+  av_dict_free(&metadata);
   for (i = 0; i < len; i++) {
 
     pair = Field(_tags, i);
@@ -1671,10 +1692,15 @@ static stream_t *new_stream(av_t *av, const AVCodec *codec) {
   stream_t *stream =
       allocate_stream_context(av, av->format_context->nb_streams, codec);
 
-  AVStream *avstream = avformat_new_stream(av->format_context, codec);
-  if (!avstream)
+  if (!stream)
     caml_raise_out_of_memory();
 
+  AVStream *avstream = avformat_new_stream(av->format_context, codec);
+  if (!avstream) {
+    free(stream);
+    caml_raise_out_of_memory();
+  }
+
   avstream->id = av->format_context->nb_streams - 1;
 
   return stream;
@@ -1745,13 +1771,21 @@ static stream_t *new_audio_stream(av_t *av, enum AVSampleFormat sample_fmt,
   return stream;
 }
 
-CAMLprim value ocaml_av_new_stream_copy(value _av, value _params) {
-  CAMLparam2(_av, _params);
+CAMLprim value ocaml_av_new_uninitialized_stream_copy(value _av) {
+  CAMLparam1(_av);
   av_t *av = Av_val(_av);
 
   stream_t *stream = new_stream(av, NULL);
 
-  AVStream *avstream = av->format_context->streams[stream->index];
+  CAMLreturn(Val_int(stream->index));
+}
+
+CAMLprim value ocaml_av_initialize_stream_copy(value _av, value _stream_index,
+                                               value _params) {
+  CAMLparam2(_av, _params);
+  av_t *av = Av_val(_av);
+
+  AVStream *avstream = av->format_context->streams[Int_val(_stream_index)];
 
   int err =
       avcodec_parameters_copy(avstream->codecpar, CodecParameters_val(_params));
@@ -1760,7 +1794,7 @@ CAMLprim value ocaml_av_new_stream_copy(value _av, value _params) {
 
   avstream->codecpar->codec_tag = 0;
 
-  CAMLreturn(Val_int(stream->index));
+  CAMLreturn(Val_unit);
 }
 
 CAMLprim value ocaml_av_new_audio_stream(value _av, value _sample_fmt,
@@ -1814,9 +1848,6 @@ static stream_t *new_video_stream(AVBufferRef *device_ctx,
                                   AVDictionary **options) {
   stream_t *stream = new_stream(av, codec);
 
-  if (!stream)
-    return NULL;
-
   init_stream_encoder(device_ctx, frame_ctx, av, stream, options);
 
   return stream;
@@ -1879,8 +1910,6 @@ CAMLprim value ocaml_av_new_video_stream(value _device_context,
 static stream_t *new_subtitle_stream(av_t *av, const AVCodec *codec,
                                      AVDictionary **options) {
   stream_t *stream = new_stream(av, codec);
-  if (!stream)
-    return NULL;
 
   int ret = subtitle_header_default(stream->codec_context);
   if (ret < 0) {
@@ -1936,6 +1965,23 @@ CAMLprim value ocaml_av_new_subtitle_stream(value _av, value _codec,
   CAMLreturn(ans);
 }
 
+CAMLprim value ocaml_av_new_data_stream(value _av, value _codec_id,
+                                        value _time_base) {
+  CAMLparam2(_av, _time_base);
+  CAMLlocal2(ans, unused);
+  const enum AVCodecID codec_id = UnknownCodecID_val(_codec_id);
+  av_t *av = Av_val(_av);
+
+  stream_t *stream = new_stream(av, NULL);
+  AVStream *s = av->format_context->streams[stream->index];
+
+  s->time_base = rational_of_value(_time_base);
+  s->codecpar->codec_type = AVMEDIA_TYPE_DATA;
+  s->codecpar->codec_id = codec_id;
+
+  CAMLreturn(Val_int(stream->index));
+}
+
 CAMLprim value ocaml_av_write_stream_packet(value _stream, value _time_base,
                                             value _packet) {
   CAMLparam3(_stream, _time_base, _packet);
@@ -2304,7 +2350,7 @@ CAMLprim value ocaml_av_close(value _av) {
 
 CAMLprim value ocaml_av_cleanup_av(value _av) {
   CAMLparam1(_av);
-  av_t *av = Av_val(_av);
+  av_t *av = Av_base_val(_av);
 
   close_av(av);
 
